##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

"""Validation modes
"""


class MODE():
    """Validation mode enumeration
    """
    NONE = 0
    SIMPLE = 1
    STRICT = 2
    VERYSTRICT = 3
