##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from typing import Optional, Dict

Translations = Optional[Dict[str, Dict[str, str]]]
